/**************************************************************************
 * 
 * PROJECT: Task Control Management (TCM)
 *
 * FILE: agenda.cc
 *
 * ABSTRACT: Defines the task agenda, which is responsible for 
 *           sending signals, executing the actions of task tree nodes, and
 *           other TBD functions.
 *
 * EXPORTS:
 *
 * $Revision: 1.14 $
 * $Date: 1999/08/04 14:00:16 $
 * $Author: reids $
 *
 * REVISION HISTORY:
 *
 * $Log: agenda.cc,v $
 * Revision 1.14  1999/08/04 14:00:16  reids
 * Distributed version of TCM.  Provides the ability for one process to
 *  add tasks, constraints, exception handlers, etc. with another process.
 *  The distributed version is created by compiling with -DDISTRIBUTED flag,
 *   which can be gotten by "gmaking" with DISTRIBUTED=1.
 *
// Revision 1.13  99/02/26  18:05:41  da0g
// Modified Agenda::clearQueues not to exit until timePoint/timeout has passed.
// 
// Revision 1.12  98/12/16  03:08:18  reids
// Added support for "on termination" functions.
//   Also enabled tca.h and tcm.h to co-exist (needed to change values
//   of several constants).
// 
// Revision 1.11  98/10/30  11:16:31  da0g
// Added ExternalEventBypassHandler.
// 
// Revision 1.10  98/10/20  22:06:23  reids
// Don't add Start_Terminating signals if signalled node is already completed.
// 
// Revision 1.9  98/07/14  17:18:32  reids
// Fixed a bug in handling Activate_Signal.
// 
// Revision 1.8  98/05/30  12:00:38  reids
// Fixed "waitFor" if timepoint is achieved but there are still events on
//   the timer queue.
// Clear timeouts related to nodes that are completely achieved.
// 
// Revision 1.7  98/04/21  17:55:16  reids
// Fixed how "forceTimeouts" decides which signals to actually send.
// 
// Revision 1.6  98/04/21  12:45:50  reids
// Added InvokeWhen/InvokeAfter functionality, to invoke a function when a
//   particular event occurs.
// Added the DelayUntilAfter/TerminateAtAfter functionality, to delay/terminate
//   a node by waiting some msecs after a particular event has occurred.
// 
// Revision 1.5  97/12/30  13:38:37  reids
// When node is terminated, signal (and remove) any timed-event in the timer
//   queue that is for that node, regardless of when it is supposed to fire.
// 
// Revision 1.4  97/12/30  12:24:47  reids
// Added option to *not* wait for timeouts and external events.
// 
// Revision 1.3  97/12/29  17:06:12  reids
// Version that has the basic functionality needed to support TDL.
// 
// Revision 1.2  97/12/04  17:50:03  reids
// Another fairly stable version (except that monitors do not quite work)
// 
// Revision 1.1  97/11/21  14:06:22  reids
// First release of TCM -- seems to be a stable version
// 
 *
 **************************************************************************/

#include "tcmBasics.h"
#include "taskTree.h"
#include "agenda.h"
#include "external.h"
#include "tcmGlobal.h"

BOOLEAN timerCompFn (const Timed_Event &timeEvent1,
		     const Timed_Event &timeEvent2)
{
#ifdef TIMER_COMP_TRACE
  tcmMessage("timerCompFn: %ld\n", 
	     timeEvent1.getTimeout() - timeEvent2.getTimeout());
#endif
  return timeEvent1 > timeEvent2;
}

/**************************************************************************
 *
 * CLASS: Agenda
 *
 **************************************************************************/

/**************************************************************************
 *
 * FUNCTION: void removeTimer(Timed_Event const &timedEvent)
 *
 * DESCRIPTION: Remove a timed-event in the timer queue that has the same
 *              signal and toNode as the given timedEvent.
 *
 * NOTE: Removes just one such event.
 *
 **************************************************************************/

void Agenda::removeTimer(Timed_Event const &timedEvent)
{
  List_Iterate<Timed_Event> timerIter(_timerQueue);
  BOOLEAN found = FALSE;
  
  for (; timerIter && !found; timerIter++) {
    Timed_Event *queuedEvent = timerIter();
    if (queuedEvent->getSignal() == timedEvent.getSignal() &&
	queuedEvent->getNode() == timedEvent.getNode()) {
      timerIter.removeCurrent();
      found = TRUE;
    }
  }
  if (!found) 
    tcmWarning("removeTimer: %s not found for %s\n",
	       signalName(timedEvent.getSignal()),
	       timedEvent.getNode()->getName());
}

/**************************************************************************
 *
 * FUNCTION: void clearQueues (Time_Point *timePoint, MSecs absoluteTime)
 *
 * DESCRIPTION: Clear the event and task queues, sending signals and 
 *              executing the actions associated with tasks.
 *
 * NOTE: Return from processing if either the time point has past or
 *       if the absoluteTime has past.
 **************************************************************************/

void Agenda::clearQueues (Time_Point *timePoint, MSecs timeout,
			  BOOLEAN waitingAllowed)
{
  struct timeval notime = {0, 0};

  if (_doneProcessing(timePoint, timeout)) return;
  do {
    if (_externalEvents.hasEvents()) {
      _externalEvents.dispatchEvents(&notime);
    }
    _clearTimerQueue();
    _clearEventQueue();
    if (_doneProcessing(timePoint, timeout)) return;
  } while (_popTaskQueue() || (waitingAllowed && _waitForEvents(timeout)) ||
	   // Not done processing, but still something to wait for
	   ((timePoint != NULL) || (timeout != INFINITE_TIME)));
}

/**************************************************************************
 *
 * FUNCTION: void clearEventQueue (void)
 *
 * DESCRIPTION: Send signals to nodes for all pending events on the even queue.
 *
 **************************************************************************/

void Agenda::_clearEventQueue (void)
{
  while (!_eventQueue.empty()) {
    Pending_Event pendingEvent = *_eventQueue.front();
    _eventQueue.pop();
    pendingEvent.invokeSignal();
  }
}

/**************************************************************************
 *
 * FUNCTION: void clearTimerQueue (void)
 *
 * DESCRIPTION: Send signals to nodes for all timed events whose timeout
 *                is now (or has past).
 *
 **************************************************************************/

void Agenda::_clearTimerQueue (void)
{
  if (!_timerQueue.empty()) {
    MSecs now = timeInMsecs();

    while (!_timerQueue.empty() && (_timerQueue.top()->getTimeout() <= now)) {
#ifdef TRACE_TIMER
      if (_timerQueue.top()->isFunctionalInvocation()) {
	tcmMessage("Signalling Functional Timed Event <%#X>\n",
		   _timerQueue.top()->getCallback());
      } else {
	tcmMessage("Signalling Timed Event <%s, %s>\n",
		   signalName(_timerQueue.top()->getSignal()),
		   _timerQueue.top()->getNode()->getName());
      }
#endif
      _timerQueue.top()->sendSignal();
      _timerQueue.pop();
    }
  }
}

/**************************************************************************
 *
 * FUNCTION: BOOLEAN popTaskQueue (void)
 *
 * DESCRIPTION: Execute the action associated with the first task tree node
 *              on the agenda's task queue (or delete the node if it is in
 *              the terminated state, and has no other references).
 *              Return TRUE if there was such a node.
 *
 **************************************************************************/

BOOLEAN Agenda::_popTaskQueue (void)
{
  if (_taskQueue.empty()) {
    return FALSE;
  } else {
    Task_Tree_Ref const task = *_taskQueue.front();
    _taskQueue.pop();
    // Execute the node, unless it is in the process of being terminated
    //  (or if it is in the "delay termination" state, used by the 
    //  "on termination" nodes
    if (task->isCurrentState(Alive_State) || task->isTerminationDelayed()) {
      task->execute();
    } else if (!task->validState(Handled_State)) {
      task->doneHandling();
    }
    return TRUE;
  }
}

BOOLEAN Agenda::_waitForEvents (MSecs timeout)
{
  // Conditions under which one should wait for a timeout (or other event)
  BOOLEAN waiting = (_taskQueue.empty() && _eventQueue.empty() &&
		     (_externalEvents.hasEvents() || !_timerQueue.empty() ||
		      timeout != INFINITE_TIME));

  if (waiting) {
    _externalEvents.dispatchEvents(_selectTimeout(timeout));
  }
  return waiting;
}

struct timeval *Agenda::_selectTimeout (MSecs maxTime)
{
  static struct timeval timeout;
  MSecs time = (_timerQueue.empty() ? INFINITE_TIME :
		_timerQueue.top()->getTimeout());
  if (maxTime < time) time = maxTime;

  if (time == INFINITE_TIME) {
    return NULL;
  } else {
    MSecs now = timeInMsecs();
    time = (time > now ? time - now : 0);
    timeout.tv_sec = time/1000;
    timeout.tv_usec = (time%1000)*1000;
    return &timeout;
  }
}

/****************************************************************
 *
 * Public Functions
 *
 *****************************************************************/

inline void enqueueEvent (Timed_Event event)
{
  GET_TCM_GLOBAL(agenda).queueTimer(event);
}

void addTimedEvent (MSecs absoluteTime,
		    const Task_Tree_Ref nodeRef, Signal_Enum signal)
{
  Timed_Event timedEvent(signal, nodeRef, absoluteTime);

  enqueueEvent(timedEvent);
  nodeRef->addExpectedEvent(signal, nodeRef, TIMER_DATA);
}

void addTimedEvent (MSecs absoluteTime,
		    SIMPLE_CALLBACK_FN_TYPE callbackFn, const void *data)
{
  Timed_Event timedEvent(callbackFn, data, absoluteTime);

  enqueueEvent(timedEvent);
}
