/**************************************************************************
 * 
 * PROJECT: Task Control Management.
 *
 * (c) Copyright 1999 Reid Simmons.  All rights reserved.
 *
 * FILE: distributed.cc
 *
 * ABSTRACT: Functions for creating distributed task tree nodes using
 *           message-passing protocols.
 *
 * $Revision: 1.2 $
 * $Date: 1999/08/05 17:22:34 $
 * $Author: reids $
 *
 * REVISION HISTORY:
 *
 * $Log: distributed.cc,v $
 * Revision 1.2  1999/08/05 17:22:34  reids
 * Changes needed because the API to the commInterface library was updated
 *   to be more generally applicable.  Removed all dependencies (at least in
 *   theory) on IPC.
 *
 * Revision 1.1  1999/08/04 14:00:17  reids
 * Distributed version of TCM.  Provides the ability for one process to
 *  add tasks, constraints, exception handlers, etc. with another process.
 *  The distributed version is created by compiling with -DDISTRIBUTED flag,
 *   which can be gotten by "gmaking" with DISTRIBUTED=1.
 *
 **************************************************************************/

#include "taskTree.h"
#include "tcmGlobal.h"
#include "virtual.h"
#include "distributed.h"
#include "commInterface.h"
#include "exception.h"

static STRING thisAgentName = "agent1";

static Distributed_Task_List distributedTasks;

/**************************************************************************
 * 
 * Local Functions
 *
 **************************************************************************/

static Distributed_Task *distributedTaskRetrieve (STRING taskName)
{
  Const_Distributed_Task_Iterator iter(distributedTasks);

  for ( ; iter; iter++) {
    if (streq((*iter()).taskName, taskName)) {
      return &(*iter());
    }
  }
  return NULL;
}

static STRING setActionMsgName (STRING taskName)
{
  static char msgName[100];

  sprintf(msgName, "%s%s", SET_ACTION_MSG_PREFIX, taskName);
  return msgName;
}

static STRING taskNameFromSetActionMsgName (STRING setActionMsgName)
{
  return &setActionMsgName[strlen(SET_ACTION_MSG_PREFIX)];
}

static STRING actionMsgFormat (STRING taskDataFormat)
{
  static char actionFormat[1000];

  sprintf(actionFormat, "{%s, *%s}", SET_ACTION_FORMAT_PREFIX,
	  (taskDataFormat ? taskDataFormat : "int"));
  return actionFormat;
}

static STRING exceptionMsgName (STRING exceptionName)
{
  static char msgName[100];

  sprintf(msgName, "%s%s", EXCEPTION_MSG_PREFIX, exceptionName);
  return msgName;
}

static STRING agentConnectionName (STRING agentName)
{
  static char connectionName[200];

  sprintf(connectionName, "%s-to-%s", thisAgent(), agentName);
  return connectionName;
}

inline int agentConnectionIndex (STRING agentName)
{
  return CommInterface::getConnectionIndex((streq(agentName, thisAgent())
					    ? agentName
					    : agentConnectionName(agentName)));
}

/**************************************************************************
 * 
 * Message-Passing Functions
 *
 **************************************************************************/

static void allocateTaskHnd (MESSAGE_REFERENCE ref, void *data)
{
  ALLOCATE_MSG_PTR msgData;     
  Distributed_Task *distributedTask;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  distributedTask = distributedTaskRetrieve(msgData->taskName);
  if (!distributedTask) {
    tcmError("%s is not registered as a distributed task\n",msgData->taskName);
  } else {
    TCM_Task_Tree_Ref node = (*distributedTask->allocationFn)();
    node->setVirtualAddress(msgData->vaddress);
  }
  CommInterface::freeData(ref, data, (void **)&msgData);
}

static void setActionHnd (MESSAGE_REFERENCE ref, void *data)
{
  SET_ACTION_MSG_PTR msgData;
  Distributed_Task *distributedTask;
  STRING taskName;

  taskName = taskNameFromSetActionMsgName(CommInterface::getRefName(ref));
  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  distributedTask = distributedTaskRetrieve(taskName);
  if (!distributedTask) {
    tcmError("%s is not registered as a distributed task\n", taskName);
  } else {
    TCM_Task_Tree_Ref node = Task_Tree_Node::getVirtual(msgData->vaddress);
    TCM_SetAction(node,
		  (*distributedTask->actionFn)(node, (void *)msgData->args));
  }
  CommInterface::freeData(ref, data, (void **)&msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Send a message to the child node's "host" telling it to add
 *              parentNode as its parent.
 *              Assumes childNode is a virtual node and parentNode already
 *              has a virtual address.
 *
 **************************************************************************/
void distributedAddParent (const Task_Tree_Ref &childNode,
			   const Task_Tree_Ref &parentNode, BOOLEAN canBeLast)
{
  ADD_PARENT_MSG_TYPE msgData;

  msgData.parentAddress = parentNode->getVirtualAddress();
  msgData.childAddress = childNode->getVirtualAddress();
  msgData.canBeLast = canBeLast;
  CommInterface::sendMessage(NODE_HOST_INDEX(childNode),
			     ADD_PARENT_MSG, &msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Handler for the "add parent" message.
 *
 **************************************************************************/
static void addParentHnd (MESSAGE_REFERENCE ref, void *data)
{
  ADD_PARENT_MSG_PTR msgData;
  Task_Tree_Ref parentNode, childNode;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  parentNode = Task_Tree_Node::getVirtual(msgData->parentAddress);
  childNode = Task_Tree_Node::getVirtual(msgData->childAddress);
  childNode->addParent(parentNode, msgData->canBeLast);
  CommInterface::freeData(ref, data, (void **)&msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Send a message to the parent node's "host" telling it to add
 *              childNode as one of its children.
 *              Assumes parentNode is a virtual node and childNode already
 *              has a virtual address.
 *
 **************************************************************************/
void distributedAddChild (const Task_Tree_Ref &parentNode,
			  const Task_Tree_Ref &childNode)
{
  ADD_CHILD_MSG_TYPE msgData;

  msgData.parentAddress = parentNode->getVirtualAddress();
  msgData.childAddress = childNode->getVirtualAddress();
  CommInterface::sendMessage(NODE_HOST_INDEX(parentNode),
			     ADD_CHILD_MSG, &msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Handler for the "add child" message.
 *
 **************************************************************************/
static void addChildHnd (MESSAGE_REFERENCE ref, void *data)
{
  ADD_CHILD_MSG_PTR msgData;
  Task_Tree_Ref parentNode, childNode;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  parentNode = Task_Tree_Node::getVirtual(msgData->parentAddress);
  childNode = Task_Tree_Node::getVirtual(msgData->childAddress);
  parentNode->addChild(childNode);
  CommInterface::freeData(ref, data, (void **)&msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Send a message to the parent node's "host" telling it to 
 *              remove childNode from its list of children.
 *              Assumes parentNode is a virtual node and childNode already
 *              has a virtual address.
 *
 **************************************************************************/
void distributedRemoveChild (const Task_Tree_Ref &parentNode,
			     const Task_Tree_Ref &childNode)
{
  REMOVE_CHILD_MSG_TYPE msgData;

  msgData.parentAddress = parentNode->getVirtualAddress();
  msgData.childAddress = childNode->getVirtualAddress();
  CommInterface::sendMessage(NODE_HOST_INDEX(parentNode),
			     REMOVE_CHILD_MSG, &msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Handler for the "remove child" message.
 *
 **************************************************************************/
static void removeChildHnd (MESSAGE_REFERENCE ref, void *data)
{
  REMOVE_CHILD_MSG_PTR msgData;
  Task_Tree_Ref parentNode, childNode;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  parentNode = Task_Tree_Node::getVirtual(msgData->parentAddress);
  childNode = Task_Tree_Node::getVirtual(msgData->childAddress);
  parentNode->removeChild(childNode);
  CommInterface::freeData(ref, data, (void **)&msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Send a message to the signalled node telling it to add
 *              <signal, signallingNode> as an expected event.
 *              Assumes signalledNode is a virtual node and signallingNode
 *              already has a virtual address.
 *
 **************************************************************************/
void distributedAddExpectedEvent (const Task_Tree_Ref &signalledNode,
				  Signal_Enum signal,
				  const Task_Tree_Ref &signallingNode,
				  const void *timerData)
{
  ADD_EXPECTED_MSG_TYPE msgData;

  msgData.signalledAddress = signalledNode->getVirtualAddress();
  msgData.signallingAddress = signallingNode->getVirtualAddress();
  msgData.signal = signal;
  msgData.timerData = (timerData ? thisAgent() : NULL);
  CommInterface::sendMessage(NODE_HOST_INDEX(signalledNode),
			     ADD_EXPECTED_MSG, &msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Handler for the "add expected event" message.
 *
 **************************************************************************/
static void addExpectedHnd (MESSAGE_REFERENCE ref, void *data)
{
  ADD_EXPECTED_MSG_PTR msgData;
  Task_Tree_Ref signalledNode, signallingNode;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  signalledNode = Task_Tree_Node::getVirtual(msgData->signalledAddress);
  signallingNode = Task_Tree_Node::getVirtual(msgData->signallingAddress);
  signalledNode->addExpectedEvent(msgData->signal, signallingNode,
				  msgData->timerData);
  CommInterface::freeData(ref, data, (void **)&msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Send a message to the signalled node signalling the event
 *              <signal, signallingNode>.
 *              Assumes signalledNode is a virtual node and signallingNode
 *              already has a virtual address.
 *
 **************************************************************************/
void distributedSignal (const Task_Tree_Ref &signalledNode, const Event &event)
{
  SIGNAL_MSG_TYPE msgData;

  msgData.signalledAddress = signalledNode->getVirtualAddress();
  msgData.signallingAddress = event.getNode()->getVirtualAddress();
  msgData.signal = event.getSignal();
  CommInterface::sendMessage(NODE_HOST_INDEX(signalledNode),
			     SIGNAL_MSG, &msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Handler for the "add expected event" message.
 *
 **************************************************************************/
static void signalHnd (MESSAGE_REFERENCE ref, void *data)
{
  SIGNAL_MSG_PTR msgData;
  Task_Tree_Ref signalledNode, signallingNode;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  signalledNode = Task_Tree_Node::getVirtual(msgData->signalledAddress);
  signallingNode = Task_Tree_Node::getVirtual(msgData->signallingAddress);

  Event event(msgData->signal, signallingNode);
  signalledNode->signal(event);
  CommInterface::freeData(ref, data, (void **)&msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Send a message to the signalling node telling it to add
 *              <state, signal, signalledNode> as a requested event.
 *              Assumes signallingNode is a virtual node and signalledNode
 *              already has a virtual address.
 *
 **************************************************************************/
void distributedAddRequestedEvent (const Task_Tree_Ref &signallingNode,
				   State_Enum state, Signal_Enum signal,
				   const Task_Tree_Ref &signalledNode)
{
  ADD_REQUESTED_MSG_TYPE msgData;

  msgData.signallingAddress = signallingNode->getVirtualAddress();
  msgData.signalledAddress = signalledNode->getVirtualAddress();
  msgData.state = state;
  msgData.signal = signal;
  CommInterface::sendMessage(NODE_HOST_INDEX(signallingNode),
			     ADD_REQUESTED_MSG, &msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Handler for the "add requested event" message.
 *
 **************************************************************************/
static void addRequestedHnd (MESSAGE_REFERENCE ref, void *data)
{
  ADD_REQUESTED_MSG_PTR msgData;
  Task_Tree_Ref signalledNode, signallingNode;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  signallingNode = Task_Tree_Node::getVirtual(msgData->signallingAddress);
  signalledNode = Task_Tree_Node::getVirtual(msgData->signalledAddress);
  signallingNode->addRequestedEvent(msgData->state, msgData->signal,
				    signalledNode);
  CommInterface::freeData(ref, data, (void **)&msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Send a message to the signalling node telling it to add
 *              a distributed callback event which will trigger when "state"
 *              holds.  Note that the callback function resides on a
 *              different process from the virtual node!
 *              Assumes signallingNode is a virtual node.
 *
 **************************************************************************/
void distributedAddRequestedCallback (const Task_Tree_Ref &signallingNode,
				      State_Enum state)
{
  ADD_REQUESTED_CALLBACK_MSG_TYPE msgData;

  msgData.signallingAddress = signallingNode->getVirtualAddress();
  msgData.state = state;
  msgData.host = thisAgent();
  CommInterface::sendMessage(NODE_HOST_INDEX(signallingNode),
			     ADD_REQUESTED_CALLBACK_MSG, &msgData);
}

static void distributedCallbackFn (const void *data)
{
  ADD_REQUESTED_CALLBACK_MSG_PTR callbackData;
  Task_Tree_Ref signallingNode;

  callbackData = (ADD_REQUESTED_CALLBACK_MSG_PTR)data;
  signallingNode = Task_Tree_Node::getVirtual(callbackData->signallingAddress);
  distributedInvokeCallback(signallingNode, callbackData->state,
			    callbackData->host);
  delete callbackData;
}

/**************************************************************************
 * 
 * DESCRIPTION: Handler for the "add requested callback" message.
 *
 **************************************************************************/
static void addRequestedCallbackHnd (MESSAGE_REFERENCE ref, void *data)
{
  ADD_REQUESTED_CALLBACK_MSG_PTR msgData;
  Task_Tree_Ref signallingNode;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  signallingNode = Task_Tree_Node::getVirtual(msgData->signallingAddress);
  signallingNode->addRequestedEvent(msgData->state, distributedCallbackFn,
				    msgData);
  // Message data will be freed later, when callback is invoked
}

/**************************************************************************
 * 
 * DESCRIPTION: Send a message to the signalling node telling it to add
 *              a distributed callback event which will trigger when "state"
 *              holds.  Note that the callback function resides on a
 *              different process from the virtual node!
 *              Assumes signallingNode is a virtual node.
 *
 **************************************************************************/
void distributedInvokeCallback (const Task_Tree_Ref &signallingNode,
				State_Enum state, STRING host)
{
  INVOKE_CALLBACK_MSG_TYPE msgData;

  msgData.signallingAddress = signallingNode->getVirtualAddress();
  msgData.state = state;
  CommInterface::sendMessage(agentConnectionIndex(host),
			     INVOKE_CALLBACK_MSG, &msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Handler for the "invoke callback" message.
 *
 **************************************************************************/
static void invokeCallbackHnd (MESSAGE_REFERENCE ref, void *data)
{
  INVOKE_CALLBACK_MSG_PTR msgData;
  Task_Tree_Ref signallingNode;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  signallingNode = Task_Tree_Node::getVirtual(msgData->signallingAddress);
  ((Virtual_Node *)*signallingNode)->invokeCallback(msgData->state);
  CommInterface::freeData(ref, data, (void **)&msgData);
}


/**************************************************************************
 * 
 * DESCRIPTION: Send a message to the signalling node telling it to remove
 *              the expected event that matches the given signal and
 *              signallingNode.
 *              Assumes signalledNode is a virtual node and signallingNode
 *              already has a virtual address.
 *
 **************************************************************************/
void distributedRemoveExpectedEvent (const Task_Tree_Ref &signalledNode,
				     Signal_Enum signal,
				     const Task_Tree_Ref &signallingNode)
{
  REMOVE_EXPECTED_MSG_TYPE msgData;

  msgData.signalledAddress = signalledNode->getVirtualAddress();
  msgData.signallingAddress = signallingNode->getVirtualAddress();
  msgData.signal = signal;
  CommInterface::sendMessage(NODE_HOST_INDEX(signalledNode),
			     REMOVE_EXPECTED_MSG, &msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Handler for the "remove expected events" message.
 *
 **************************************************************************/
static void removeExpectedHnd (MESSAGE_REFERENCE ref, void *data)
{
  REMOVE_EXPECTED_MSG_PTR msgData;
  Task_Tree_Ref signalledNode, signallingNode;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  signalledNode = Task_Tree_Node::getVirtual(msgData->signalledAddress);
  signallingNode = Task_Tree_Node::getVirtual(msgData->signallingAddress);
  signalledNode->removeExpectedEvent(msgData->signal, signallingNode);
  CommInterface::freeData(ref, data, (void **)&msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Send a message to the signalling node telling it to remove
 *              all the <signal, signalledNode, state> requested events
 *              that match the given signal and signalledNode.
 *              Assumes signallingNode is a virtual node and signalledNode
 *              already has a virtual address.
 *
 **************************************************************************/
void distributedRemoveRequestedEvents (const Task_Tree_Ref &signallingNode,
				       Signal_Enum signal,
				       const Task_Tree_Ref &signalledNode)
{
  REMOVE_REQUESTED_MSG_TYPE msgData;

  msgData.signallingAddress = signallingNode->getVirtualAddress();
  msgData.signalledAddress = signalledNode->getVirtualAddress();
  msgData.signal = signal;
  CommInterface::sendMessage(NODE_HOST_INDEX(signallingNode),
			     REMOVE_REQUESTED_MSG, &msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Handler for the "remove requested events" message.
 *
 **************************************************************************/
static void removeRequestedHnd (MESSAGE_REFERENCE ref, void *data)
{
  REMOVE_REQUESTED_MSG_PTR msgData;
  Task_Tree_Ref signalledNode, signallingNode;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  signallingNode = Task_Tree_Node::getVirtual(msgData->signallingAddress);
  signalledNode = Task_Tree_Node::getVirtual(msgData->signalledAddress);
  signallingNode->removeRequestedEvents(msgData->signal, signalledNode);
  CommInterface::freeData(ref, data, (void **)&msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Remove the timed event from the given host.
 *              Assumes that the node has a virtual address.
 *
 **************************************************************************/
void distributedRemoveTimer (const Task_Tree_Ref &nodeRef, Signal_Enum signal,
			     STRING timerHost)
{
  REMOVE_TIMER_MSG_TYPE msgData;

  msgData.vaddress = nodeRef->getVirtualAddress();
  msgData.signal = signal;
  CommInterface::sendMessage(agentConnectionIndex(timerHost),
			     REMOVE_TIMER_MSG, &msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Handler for the "remove timer events" message.
 *
 **************************************************************************/
static void removeTimerHnd (MESSAGE_REFERENCE ref, void *data)
{
  REMOVE_TIMER_MSG_PTR msgData;
  Task_Tree_Ref nodeRef;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  nodeRef = Task_Tree_Node::getVirtual(msgData->vaddress);

  Timed_Event timedEvent(msgData->signal, nodeRef, 0);
  GET_TCM_GLOBAL(agenda).removeTimer(timedEvent);
  CommInterface::freeData(ref, data, (void **)&msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Send a message to the node telling it to terminate itself
 *              and its children.  Assumes nodeRef is a virtual node.
 *
 **************************************************************************/
void distributedTerminate (const Task_Tree_Ref &nodeRef)
{
  TERMINATE_MSG_TYPE msgData;

  msgData.vaddress = nodeRef->getVirtualAddress();
  CommInterface::sendMessage(NODE_HOST_INDEX(nodeRef), TERMINATE_MSG, &msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Handler for the "terminate" message.
 *
 **************************************************************************/
static void terminateHnd (MESSAGE_REFERENCE ref, void *data)
{
  TERMINATE_MSG_PTR msgData;
  Task_Tree_Ref nodeRef;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  nodeRef = Task_Tree_Node::getVirtual(msgData->vaddress);
  nodeRef->terminate();
  CommInterface::freeData(ref, data, (void **)&msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Send a message to the node telling it to deallocate itself.
 *              Assumes nodeRef is a virtual node.
 *
 **************************************************************************/
void distributedDeallocate (const Task_Tree_Ref &nodeRef)
{
  DEALLOCATE_MSG_TYPE msgData;

  msgData.vaddress = nodeRef->getVirtualAddress();
  CommInterface::sendMessage(NODE_HOST_INDEX(nodeRef), DEALLOCATE_MSG, &msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Handler for the "deallocate" message.
 *
 **************************************************************************/
static void deallocateHnd (MESSAGE_REFERENCE ref, void *data)
{
  DEALLOCATE_MSG_PTR msgData;
  Task_Tree_Ref nodeRef;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  nodeRef = Task_Tree_Node::getVirtual(msgData->vaddress);
  nodeRef->deallocate();
  CommInterface::freeData(ref, data, (void **)&msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Send a message to the node telling it to suspend itself
 *              and its children.  Assumes nodeRef is a virtual node.
 *
 **************************************************************************/
void distributedSuspend (const Task_Tree_Ref &nodeRef)
{
  SUSPEND_MSG_TYPE msgData;

  msgData.vaddress = nodeRef->getVirtualAddress();
  CommInterface::sendMessage(NODE_HOST_INDEX(nodeRef), SUSPEND_MSG, &msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Handler for the "suspend" message.
 *
 **************************************************************************/
static void suspendHnd (MESSAGE_REFERENCE ref, void *data)
{
  SUSPEND_MSG_PTR msgData;
  Task_Tree_Ref nodeRef;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  nodeRef = Task_Tree_Node::getVirtual(msgData->vaddress);
  nodeRef->suspend();
  CommInterface::freeData(ref, data, (void **)&msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Send a message to the node telling it to unsuspend itself
 *              and its children.  Assumes nodeRef is a virtual node.
 *
 **************************************************************************/
void distributedUnsuspend (const Task_Tree_Ref &nodeRef)
{
  UNSUSPEND_MSG_TYPE msgData;

  msgData.vaddress = nodeRef->getVirtualAddress();
  CommInterface::sendMessage(NODE_HOST_INDEX(nodeRef), UNSUSPEND_MSG, &msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Handler for the "unsuspend" message.
 *
 **************************************************************************/
static void unsuspendHnd (MESSAGE_REFERENCE ref, void *data)
{
  UNSUSPEND_MSG_PTR msgData;
  Task_Tree_Ref nodeRef;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  nodeRef = Task_Tree_Node::getVirtual(msgData->vaddress);
  nodeRef->unsuspend();
  CommInterface::freeData(ref, data, (void **)&msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Send a message to the node telling it to delay termination.
 *              Assumes nodeRef is a virtual node.
 *
 **************************************************************************/
void distributedDelayTermination (const Task_Tree_Ref &nodeRef)
{
  DELAY_TERMINATION_MSG_TYPE msgData;

  msgData.vaddress = nodeRef->getVirtualAddress();
  CommInterface::sendMessage(NODE_HOST_INDEX(nodeRef), DELAY_TERMINATION_MSG,
			     &msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Handler for the "delayTermination" message.
 *
 **************************************************************************/
static void delayTerminationHnd (MESSAGE_REFERENCE ref, void *data)
{
  DELAY_TERMINATION_MSG_PTR msgData;
  Task_Tree_Ref nodeRef;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  nodeRef = Task_Tree_Node::getVirtual(msgData->vaddress);
  nodeRef->delayTermination();
  CommInterface::freeData(ref, data, (void **)&msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Send a message to the node telling it to enable termination.
 *              Assumes nodeRef is a virtual node.
 *
 **************************************************************************/
void distributedEnableTermination (const Task_Tree_Ref &nodeRef)
{
  ENABLE_TERMINATION_MSG_TYPE msgData;

  msgData.vaddress = nodeRef->getVirtualAddress();
  CommInterface::sendMessage(NODE_HOST_INDEX(nodeRef), ENABLE_TERMINATION_MSG,
			     &msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Handler for the "enableTermination" message.
 *
 **************************************************************************/
static void enableTerminationHnd (MESSAGE_REFERENCE ref, void *data)
{
  ENABLE_TERMINATION_MSG_PTR msgData;
  Task_Tree_Ref nodeRef;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  nodeRef = Task_Tree_Node::getVirtual(msgData->vaddress);
  nodeRef->enableTermination();
  CommInterface::freeData(ref, data, (void **)&msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Send a message to the node asking it to display its subtree.
 *              Assumes nodeRef is a virtual node.
 *
 * NOTES: Can only display to either stdout or stderr (whereas the regular
 *        version can display to any arbitrary stream/file).
 *
 **************************************************************************/
void distributedDisplayTree (const Task_Tree_Ref &nodeRef,
			     int whichStream, int position)
{
  DISPLAY_TREE_MSG_TYPE msgData;

  msgData.vaddress = nodeRef->getVirtualAddress();
  msgData.whichStream = whichStream;
  msgData.position = position;
  CommInterface::sendMessage(NODE_HOST_INDEX(nodeRef),
			     DISPLAY_TREE_MSG, &msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Handler for the "display tree" message.
 *
 **************************************************************************/
static void displayTreeHnd (MESSAGE_REFERENCE ref, void *data)
{
  DISPLAY_TREE_MSG_PTR msgData;
  Task_Tree_Ref nodeRef;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  nodeRef = Task_Tree_Node::getVirtual(msgData->vaddress);
  nodeRef->displayTree((msgData->whichStream == USE_STDOUT ? stdout : stderr),
		       msgData->position);
  CommInterface::freeData(ref, data, (void **)&msgData);
}


/**************************************************************************
 * 
 * DESCRIPTION: Send a message to the node's "host" telling it to add
 *              terminationRef as a new termination action
 *              Assumes nodeRef is a virtual node and terminationRef
 *              already has a virtual address.
 *
 **************************************************************************/
void distributedAddTermination (const Task_Tree_Ref &nodeRef,
				const Task_Tree_Ref &terminationNode)
{
  ADD_TERMINATION_MSG_TYPE msgData;

  msgData.vaddress = nodeRef->getVirtualAddress();
  msgData.terminationAddress = terminationNode->getVirtualAddress();
  msgData.terminationName = TCM_NodeName(terminationNode);
  CommInterface::sendMessage(NODE_HOST_INDEX(nodeRef),
			     ADD_TERMINATION_MSG, &msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Handler for the "add termination action" message.
 *
 **************************************************************************/
static void addTerminationHnd (MESSAGE_REFERENCE ref, void *data)
{
  ADD_TERMINATION_MSG_PTR msgData;
  Task_Tree_Ref nodeRef, terminationNode;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  nodeRef = Task_Tree_Node::getVirtual(msgData->vaddress);
  terminationNode = Task_Tree_Node::getVirtual(msgData->terminationAddress,
					       msgData->terminationName);
  nodeRef->addTerminationAction(terminationNode);
  CommInterface::freeData(ref, data, (void **)&msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Send a message to the node asking it find an exception
 *              handler for the failure and add an exception node.
 *              Assumes nodeRef is a virtual node and parentNode and
 *              failureNode already have virtual addresses.
 *
 **************************************************************************/

void distributedFindAndApplyExceptionHandler (const Task_Tree_Ref &nodeRef,
					      const Task_Tree_Ref &parentNode,
					      const Task_Tree_Ref &failureNode,
					      Exception_Instance *exception,
					      const void *data,
					      BOOLEAN isRefCountFailureData)
{
  FIND_EXCEPTION_HANDLER_MSG_TYPE msgData;

  msgData.contextNode = nodeRef->getVirtualAddress();
  msgData.parentNode =  parentNode->getVirtualAddress();
  msgData.failureNode = failureNode->getVirtualAddress();
  msgData.failure = exception->getName();

  bzero(&msgData.exceptionData, sizeof(msgData.exceptionData));
  msgData.exceptionData.isRefCount = isRefCountFailureData;
  if (data) {
    unsigned length;
    if (CommInterface::marshallData(exceptionMsgName(msgData.failure),
				    (void *)data, &msgData.exceptionData.data,
				    &length) == SUCCESS) {
      msgData.exceptionData.length = length;
    } else {
      tcmError("Need to call TCM_RegisterDistributedException for '%s'\n",
	       msgData.failure);
    }
  }

  CommInterface::sendMessage(NODE_HOST_INDEX(nodeRef),
			     FIND_EXCEPTION_HANDLER_MSG, &msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Handler for the "find and apply exception handler" message.
 *
 **************************************************************************/
static void findAndApplyExceptionHandlerHnd (MESSAGE_REFERENCE ref, void *data)
{
  FIND_EXCEPTION_HANDLER_MSG_PTR msgData;
  Task_Tree_Ref contextNode, parentNode, failureNode;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  contextNode = Task_Tree_Node::getVirtual(msgData->contextNode);
  parentNode  = Task_Tree_Node::getVirtual(msgData->parentNode);
  failureNode = Task_Tree_Node::getVirtual(msgData->failureNode);
  Exception_Instance exception(msgData->failure, contextNode);
  if (msgData->exceptionData.length == 0) {
    contextNode->findAndApplyExceptionHandler(parentNode, failureNode,
					      &exception, NULL, FALSE);
  } else {
    void *dataPtr, *exceptionData;
    BOOLEAN isRefCount = msgData->exceptionData.isRefCount;

    if (CommInterface::unmarshallData(exceptionMsgName(msgData->failure),
				      msgData->exceptionData.data,
				      &dataPtr) == SUCCESS) {
      if (isRefCount) {
	tcmError("Cannot handle passing ref_count data for exception '%s'\n",
		 msgData->failure);
      } else {
	// Ugh -- but need to force type coercion
	exceptionData = (void *)*(int *)dataPtr;
      }
      contextNode->findAndApplyExceptionHandler(parentNode, failureNode,
						&exception, exceptionData,
						isRefCount);
    } else {
      tcmError("Need to call TCM_RegisterDistributedException for '%s'\n",
	       msgData->failure);
    }
  }
  CommInterface::freeData(ref, data, (void **)&msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Send a message to the node telling it that no exception 
 *              handler was found for the failure.
 *              Assumes nodeRef is a virtual node.
 *
 **************************************************************************/
void distributedNoExceptionHandler (const Task_Tree_Ref &nodeRef,
				    STRING failure)
{
  FIND_EXCEPTION_HANDLER_MSG_TYPE msgData;

  msgData.failureNode = nodeRef->getVirtualAddress();
  msgData.failure = failure;
  CommInterface::sendMessage(NODE_HOST_INDEX(nodeRef),
			     FIND_EXCEPTION_HANDLER_MSG, &msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Handler for the "no exception handler" message.
 *
 **************************************************************************/
static void noExceptionHandlerHnd (MESSAGE_REFERENCE ref, void *data)
{
  FIND_EXCEPTION_HANDLER_MSG_PTR msgData;
  Task_Tree_Ref nodeRef;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  nodeRef = Task_Tree_Node::getVirtual(msgData->failureNode);
  nodeRef->noExceptionHandler(msgData->failure);
  CommInterface::freeData(ref, data, (void **)&msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Send a message to the node asking it to check whether
 *              the timed event should be added, and if not to dequeue
 *              it from this process.
 *              Assumes afterRef is a virtual node, nodeRef has the
 *              virtual address set, and the timed event has already been
 *              added.
 *
 **************************************************************************/
void distributedCheckTimedEventAfterSignal (const Task_Tree_Ref afterRef, 
					    const Task_Tree_Ref nodeRef, 
					    Signal_Enum signal,
					    State_Enum state)
{
  CHECK_TIMED_EVENT_MSG_TYPE msgData;

  msgData.afterNode = afterRef->getVirtualAddress();
  msgData.nodeRef = nodeRef->getVirtualAddress();
  msgData.state = state;
  msgData.signal = signal;
  msgData.host = thisAgent();
  CommInterface::sendMessage(NODE_HOST_INDEX(afterRef),
			     CHECK_TIMED_EVENT_MSG, &msgData);
}

/**************************************************************************
 * 
 * DESCRIPTION: Handler for the "check timedd event" message.
 *
 **************************************************************************/
static void checkTimedEventAfterSignalHnd (MESSAGE_REFERENCE ref, void *data)
{
  CHECK_TIMED_EVENT_MSG_PTR msgData;
  Task_Tree_Ref afterNode, nodeRef;

  CommInterface::unmarshallData(ref, data, (void **)&msgData);
  afterNode = Task_Tree_Node::getVirtual(msgData->afterNode);
  nodeRef = Task_Tree_Node::getVirtual(msgData->nodeRef);

  if (!ADD_EVENT_P(msgData->signal, afterNode, msgData->state)) {
    nodeRef->removeExpectedEvent(msgData->signal, nodeRef);
    distributedRemoveTimer(nodeRef, msgData->signal, msgData->host);
#ifdef TRACE_AFTER_SIGNAL
    tcmWarning("ASF: %s %s <%s, %s, %s>\n", 
	       afterNode->getName(), 
	       stateName(afterNode->theCurrentState(Alive_State)),
	       signalName(msgData->signal), nodeRef->getName(), 
	       stateName(msgData->state));
#endif
  }

  CommInterface::freeData(ref, data, (void **)&msgData);
}

static void registerDistributedMessages (void)
{
  CommInterface::registerMessage(ALLOCATE_MSG, ALLOCATE_MSG_FORMAT);
  CommInterface::subscribeMessage(ALLOCATE_MSG, allocateTaskHnd);

  CommInterface::registerMessage(ADD_PARENT_MSG, ADD_PARENT_MSG_FORMAT);
  CommInterface::subscribeMessage(ADD_PARENT_MSG, addParentHnd);

  CommInterface::registerMessage(ADD_CHILD_MSG, ADD_CHILD_MSG_FORMAT);
  CommInterface::subscribeMessage(ADD_CHILD_MSG, addChildHnd);

  CommInterface::registerMessage(REMOVE_CHILD_MSG, REMOVE_CHILD_MSG_FORMAT);
  CommInterface::subscribeMessage(REMOVE_CHILD_MSG, removeChildHnd);

  CommInterface::registerMessage(SIGNAL_MSG, SIGNAL_MSG_FORMAT);
  CommInterface::subscribeMessage(SIGNAL_MSG, signalHnd);

  CommInterface::registerMessage(ADD_EXPECTED_MSG, ADD_EXPECTED_MSG_FORMAT);
  CommInterface::subscribeMessage(ADD_EXPECTED_MSG, addExpectedHnd);

  CommInterface::registerMessage(ADD_REQUESTED_MSG, ADD_REQUESTED_MSG_FORMAT);
  CommInterface::subscribeMessage(ADD_REQUESTED_MSG, addRequestedHnd);

  CommInterface::registerMessage(ADD_REQUESTED_CALLBACK_MSG,
				 ADD_REQUESTED_CALLBACK_MSG_FORMAT);
  CommInterface::subscribeMessage(ADD_REQUESTED_CALLBACK_MSG,
				  addRequestedCallbackHnd);

  CommInterface::registerMessage(INVOKE_CALLBACK_MSG,
				 INVOKE_CALLBACK_MSG_FORMAT);
  CommInterface::subscribeMessage(INVOKE_CALLBACK_MSG, invokeCallbackHnd);

  CommInterface::registerMessage(REMOVE_EXPECTED_MSG,
				 REMOVE_EXPECTED_MSG_FORMAT);
  CommInterface::subscribeMessage(REMOVE_EXPECTED_MSG, removeExpectedHnd);

  CommInterface::registerMessage(REMOVE_REQUESTED_MSG,
				 REMOVE_REQUESTED_MSG_FORMAT);
  CommInterface::subscribeMessage(REMOVE_REQUESTED_MSG, removeRequestedHnd);

  CommInterface::registerMessage(REMOVE_TIMER_MSG, REMOVE_TIMER_MSG_FORMAT);
  CommInterface::subscribeMessage(REMOVE_TIMER_MSG, removeTimerHnd);

  CommInterface::registerMessage(TERMINATE_MSG, TERMINATE_MSG_FORMAT);
  CommInterface::subscribeMessage(TERMINATE_MSG, terminateHnd);

  CommInterface::registerMessage(DEALLOCATE_MSG, DEALLOCATE_MSG_FORMAT);
  CommInterface::subscribeMessage(DEALLOCATE_MSG, deallocateHnd);

  CommInterface::registerMessage(SUSPEND_MSG, SUSPEND_MSG_FORMAT);
  CommInterface::subscribeMessage(SUSPEND_MSG, suspendHnd);

  CommInterface::registerMessage(UNSUSPEND_MSG, UNSUSPEND_MSG_FORMAT);
  CommInterface::subscribeMessage(UNSUSPEND_MSG, unsuspendHnd);

  CommInterface::registerMessage(DELAY_TERMINATION_MSG,
				 DELAY_TERMINATION_MSG_FORMAT);
  CommInterface::subscribeMessage(DELAY_TERMINATION_MSG, delayTerminationHnd);

  CommInterface::registerMessage(ENABLE_TERMINATION_MSG,
				 ENABLE_TERMINATION_MSG_FORMAT);
  CommInterface::subscribeMessage(ENABLE_TERMINATION_MSG, 
				  enableTerminationHnd);

  CommInterface::registerMessage(ADD_TERMINATION_MSG,
				 ADD_TERMINATION_MSG_FORMAT);
  CommInterface::subscribeMessage(ADD_TERMINATION_MSG, addTerminationHnd);

  CommInterface::registerMessage(DISPLAY_TREE_MSG, DISPLAY_TREE_MSG_FORMAT);
  CommInterface::subscribeMessage(DISPLAY_TREE_MSG, displayTreeHnd);

  CommInterface::registerMessage(FIND_EXCEPTION_HANDLER_MSG,
				 FIND_EXCEPTION_HANDLER_MSG_FORMAT);
  CommInterface::subscribeMessage(FIND_EXCEPTION_HANDLER_MSG, 
				  findAndApplyExceptionHandlerHnd);

  CommInterface::registerMessage(NO_EXCEPTION_HANDLER_MSG,
				 NO_EXCEPTION_HANDLER_MSG_FORMAT);
  CommInterface::subscribeMessage(NO_EXCEPTION_HANDLER_MSG, 
				  noExceptionHandlerHnd);

  CommInterface::registerMessage(CHECK_TIMED_EVENT_MSG,
				 CHECK_TIMED_EVENT_MSG_FORMAT);
  CommInterface::subscribeMessage(CHECK_TIMED_EVENT_MSG,
				  checkTimedEventAfterSignalHnd);
}

/**************************************************************************
 * 
 * User-Interface Functions
 *
 **************************************************************************/

/**************************************************************************
 * 
 * DESCRIPTION: Return the name of the agent associated with this process.
 *
 **************************************************************************/
STRING thisAgent (void)
{
  return thisAgentName;
}

/**************************************************************************
 * 
 * DESCRIPTION: Create and return a virtual node, and send a message to
 *              the "agentName" to create a node of "taskName".
 *
 **************************************************************************/
TCM_Task_Tree_Ref TCM_AllocateDistributedNode (STRING agentName,
					       STRING taskName)
{
  Task_Tree_Ref virtualNode;
  ALLOCATE_MSG_TYPE msgData;

  if (!CommInterface::getCommHasBeenInitialized()) {
    tcmError("Need to call TCM_EnableDistributedComm first\n");
  } else if (streq(agentName, thisAgent())) {
    tcmError("Trying to allocate a distributed node (%s) internally\n");
  } else {
    virtualNode = new Virtual_Node(taskName, agentName, thisAgent());
    msgData.vaddress = virtualNode->getVirtualAddress();
    msgData.taskName = taskName;
    CommInterface::sendMessage(agentConnectionIndex(agentName),
			       ALLOCATE_MSG, &msgData);
  }
  return virtualNode;
}

/**************************************************************************
 * 
 * DESCRIPTION: Send a message to the virtual address of the node to
 *              set the action with the given data.
 *
 **************************************************************************/
TCM_Return_Type TCM_SetDistributedAction (const TCM_Task_Tree_Ref &nodeRef, 
					  const void *callbackData)
{
  SET_ACTION_MSG_TYPE msgData;
  STRING agentName;

  if (!IS_DISTRIBUTED(nodeRef)) {
    tcmError("TCM_SetDistributedAction: %s is not a distributed node\n",
	     nodeRef->instanceName());
    return TCM_Error;
  } else if (!CommInterface::getCommHasBeenInitialized()) {
    tcmError("Need to call TCM_EnableDistributedComm first\n");
    return TCM_Error;
  } else {
    msgData.vaddress = nodeRef->getVirtualAddress();
    msgData.args = callbackData;
    agentName = nodeRef->getVirtualAddress().host;
    CommInterface::sendMessage(agentConnectionIndex(agentName),
			       setActionMsgName(nodeRef->instanceName()),
			       &msgData);
    return TCM_Ok;
  }
}

/**************************************************************************
 * 
 * DESCRIPTION: This is a shorthand for a series of allocate/setAction/insert
 *              calls
 *
 **************************************************************************/
TCM_Task_Tree_Ref TCM_CreateDistributedNode(STRING agentName,
					    const TCM_Task_Tree_Ref &parentRef,
					    STRING taskName,
					    void *callbackData,
					    BOOLEAN isLastChild)
{
  TCM_Task_Tree_Ref nodeRef;
  if (!validRef(parentRef)) {
    tcmWarning("TCM_CreateDistributedNode: Reference to NULL node\n");
  } else if (!CommInterface::getCommHasBeenInitialized()) {
    tcmError("Need to call TCM_EnableDistributedComm first\n");
  } else if (streq(agentName, thisAgent())) {
    tcmError("Trying to allocate a distributed node (%s) internally\n");
  } else {
    nodeRef = TCM_AllocateDistributedNode(agentName, taskName);
    TCM_SetDistributedAction(nodeRef, callbackData);
    taskTreeInsert(*parentRef, *nodeRef, isLastChild);
  }
  return nodeRef;
}

/**************************************************************************
 * 
 * DESCRIPTION: Register "taskName" as a task that can be invoked by
 *              another module (using "CreateDistributedNode").
 *
 **************************************************************************/
TCM_Return_Type TCM_RegisterDistributedTask (STRING taskName, 
					     TASK_ALLOCATION_FN allocationFn,
					     ACTION_CREATION_FN actionFn,
					     STRING taskDataFormat)
{
  Distributed_Task distributedTask, *existingTask;
  STRING actionMsgName;

  if (!CommInterface::getCommHasBeenInitialized()) {
    tcmError("Need to call TCM_EnableDistributedComm first\n");
    return TCM_Error;
  } else {
    existingTask = distributedTaskRetrieve(taskName);
    if (existingTask) {
      tcmWarning("%s is already registered as a distributed task\n", taskName);
      existingTask->allocationFn = allocationFn;
      existingTask->actionFn = actionFn;
    } else {
      distributedTask.taskName = taskName;
      distributedTask.allocationFn = allocationFn;
      distributedTask.actionFn = actionFn;
      distributedTasks.insertLast(distributedTask);
    }
    actionMsgName = setActionMsgName(taskName);
    CommInterface::registerMessage(actionMsgName,
				   actionMsgFormat(taskDataFormat));
    CommInterface::subscribeMessage(actionMsgName, setActionHnd);
    return TCM_Ok;
  }
}

TCM_Return_Type TCM_EnableDistributedComm (STRING myName, STRING myHostMachine)
{
  if (CommInterface::initializeComm(myName, myHostMachine,
				    registerDistributedMessages) == SUCCESS) {
    TCM_SetExternalEventBypassHandler(CommInterface::eventHandlerForTCM);
    thisAgentName = myName;
    return TCM_Ok;
  } else {
    return TCM_Error;
  }
}

TCM_Return_Type TCM_ConnectDistributedAgent (STRING agentName,
					     STRING hostMachine)
{
  if (!CommInterface::getCommHasBeenInitialized()) {
    tcmError("Need to call TCM_EnableDistributedComm first\n");
    return TCM_Error;
  } else if (agentConnectionIndex(agentName) >= 0) {
    tcmWarning("Already connected to agent with name %s\n", agentName);
    return TCM_Error;
  } else {
    return (CommInterface::openConnection(agentConnectionName(agentName),
					  hostMachine)
	    == SUCCESS ? TCM_Ok : TCM_Error);
  }
}

TCM_Return_Type TCM_RegisterDistributedException (STRING exceptionName, 
						  STRING exceptionDataFormat)
{
  if (!CommInterface::getCommHasBeenInitialized()) {
    tcmError("Need to call TCM_EnableDistributedComm first\n");
    return TCM_Error;
  } else {
    CommInterface::registerMessage(exceptionMsgName(exceptionName),
				   exceptionDataFormat);
    return TCM_Ok;
  }
}
