/**************************************************************************
 * 
 * PROJECT: Task Control Management.
 *
 * (c) Copyright 1996 Reid Simmons.  All rights reserved.
 *
 * FILE: event.h
 *
 * ABSTRACT: Defines signals, events (occurrence of signals), state, 
 *           and state-events (events that trigger state transitions).
 *           Main mode of communication between task tree nodes.
 *
 * EXPORTS:
 *
 * $Revision: 1.6 $
 * $Date: 1999/10/21 21:32:31 $
 * $Author: reids $
 *
 * REVISION HISTORY:
 *
 * $Log: event.h,v $
 * Revision 1.6  1999/10/21 21:32:31  reids
 * Changed list.cc and list.h to tcmList.cc and tcmList.h to avoid conflicts
 *   with other occurrences of those names in other packages.
 *
 * Revision 1.5  1999/08/04 14:00:17  reids
 * Distributed version of TCM.  Provides the ability for one process to
 *  add tasks, constraints, exception handlers, etc. with another process.
 *  The distributed version is created by compiling with -DDISTRIBUTED flag,
 *   which can be gotten by "gmaking" with DISTRIBUTED=1.
 *
 * Revision 1.4  98/06/02  10:39:34  reids
 * Increased overall efficiency of TCM.
 * 
 * Revision 1.3  98/04/21  12:45:56  reids
 * Added InvokeWhen/InvokeAfter functionality, to invoke a function when a
 *   particular event occurs.
 * Added the DelayUntilAfter/TerminateAtAfter functionality, to delay/terminate
 *   a node by waiting some msecs after a particular event has occurred.
 * 
 * Revision 1.2  97/12/04  17:50:06  reids
 * Another fairly stable version (except that monitors do not quite work)
 * 
 * Revision 1.1  97/11/21  14:06:27  reids
 * First release of TCM -- seems to be a stable version
 * 
 *
 **************************************************************************/

#ifndef INCevent
#define INCevent

#include "tcmBasics.h"
#include "tcmList.h"
#include "tcm.h"
#include "tcmPriv.h"

enum Signal_Enum { Unknown_Signal,
		   Start_Handling_Signal, Done_Handling_Signal,
		   Start_Planning_Signal, Done_Planning_Signal,
		   Start_Achieving_Signal, Done_Achieving_Signal,
		   Start_Terminating_Signal, Done_Terminating_Signal,
		   Activate_Signal, Timeout_Signal};

enum State_Enum { Unknown_State, 
		  Unhandled_State, Handling_State, Handled_State,
		  Unplanned_State, Planning_State, Planned_State,
		  Unachieved_State, Achieving_State, Achieved_State,
		  Alive_State, Terminating_State, Terminated_State};

enum Phase_Enum { Handling_Phase, Planning_Phase, Achieving_Phase, 
		  Terminating_Phase};

STRING signalName (Signal_Enum signal);

STRING stateName (State_Enum state);

State_Enum signalToState(Signal_Enum signal);
Signal_Enum stateToSignal(State_Enum state);

Phase_Enum stateToPhase(State_Enum state);

class State
{
 public:
  State() { phases[Handling_Phase] = Unhandled_State;
	    phases[Achieving_Phase] = Unachieved_State;
	    phases[Planning_Phase] = Unplanned_State; 
	    phases[Terminating_Phase] = Alive_State; }
  void set (State_Enum state);
  State_Enum next (State_Enum state) const ;
  int compare (State_Enum state) const;
  State_Enum current (State_Enum state) const 
    { return phases[stateToPhase(state)]; }

 private:
  State_Enum phases[4];
};

// This is pretty kludgy, but could not think of a better way that didn't
// involve creating new objects, which I want to avoid since Event's are
// very common.
// There are two flavors of events -- "node" events signal task tree nodes;
//  "functional" events merely invoke functions.

class Invocation
{
 public:
  Invocation()
    : signal(Unknown_Signal), node(NULL), callback(NULL) { data = NULL; }
  Invocation(Signal_Enum theSignal, Task_Tree_Ref const &theNode)
    : signal(theSignal), node(theNode), callback(NULL) { data = NULL; }
  Invocation(const SIMPLE_CALLBACK_FN_TYPE callbackFn, 
	     const void *callbackData = NULL) 
    : signal(Unknown_Signal), node(NULL), callback(callbackFn) 
    { data = callbackData; }

  // Differentiate "node" events from "functional" events
  BOOLEAN isFunctionalInvocation(void) const { return callback != NULL; }

  Signal_Enum getSignal() const { return signal; }
  Task_Tree_Ref const &getNode() const { return node; }

  SIMPLE_CALLBACK_FN_TYPE getCallback() const { return callback; }
  const void *getData() const { return data; }

  BOOLEAN operator==(const Invocation &invocation1) const
    { return (isFunctionalInvocation()
	      ? (callback == invocation1.getCallback() &&
		 data == invocation1.getData())
	      : (node == invocation1.getNode() &&
		 signal == invocation1.getSignal())); }

 protected:
  // For "node" events
  Signal_Enum signal;
  Task_Tree_Ref node;
  // For "functional" events
  SIMPLE_CALLBACK_FN_TYPE callback;
  const void *data;
};

class Event : public Invocation
{
 public:
  Event(void) {}
  Event(Signal_Enum theSignal, Task_Tree_Ref const &theNode)
    : Invocation(theSignal, theNode) {}
  Event(SIMPLE_CALLBACK_FN_TYPE callbackFn, const void *callbackData)
    : Invocation(callbackFn, callbackData) { }

  void setData (const void *theData) { data = theData; }

  void sendSignal(Task_Tree_Ref const &fromNode) const;
};

class State_Event : public Event
{
 public:
  State_Event(void) { state = Unknown_State; }
  State_Event(Signal_Enum theSignal, Task_Tree_Ref const &theNode,
	      State_Enum theState)
    : Event(theSignal, theNode) { state = theState; }
  State_Event(SIMPLE_CALLBACK_FN_TYPE callbackFn, const void *callbackData,
	      State_Enum theState)
    : Event(callbackFn, callbackData) { state = theState; }

  BOOLEAN operator==(const State_Event &event1) const
    { return ((state == event1.getState()) && 
	      this->Event::operator==(event1)); }

  State_Enum getState(void) const { return state; }

 private:
  State_Enum state;
};

class Pending_Event : public Event
{
 public:
  Pending_Event() : fromNode(NULL) {}
  Pending_Event(Signal_Enum theSignal, Task_Tree_Ref const &toNode,
		Task_Tree_Ref const &theFromNode)
    : Event(theSignal, toNode), fromNode(theFromNode) {}
  Pending_Event(SIMPLE_CALLBACK_FN_TYPE callbackFn, const void *callbackData,
		Task_Tree_Ref const &theFromNode)
    : Event(callbackFn, callbackData), fromNode(theFromNode) {}

  BOOLEAN operator==(const Pending_Event &event1) const
    { return ((fromNode == event1.getFromNode()) &&
	      this->Event::operator==(event1)); }

  Task_Tree_Ref const &getFromNode() const { return fromNode; };
  Task_Tree_Ref const &getToNode() const { return getNode(); };

  void invokeSignal (void) const;

 private:
  Task_Tree_Ref fromNode;
};

/****************************************************************
 *
 * "Timed_Event"s are events that trigger at a given (absolute) time.
 * They are managed by the agenda's timer-queue (a priority queue).
 *
 *****************************************************************/

class Timed_Event : public Event
{
 public:
  Timed_Event() { _timeout = 0; }
  Timed_Event(Signal_Enum signal, Task_Tree_Ref const &node, MSecs timeout)
    : Event(signal, node) { _timeout = timeout; }
  Timed_Event(SIMPLE_CALLBACK_FN_TYPE callbackFn, const void *callbackData,
	      MSecs timeout)
    : Event(callbackFn, callbackData) { _timeout = timeout; }

  BOOLEAN operator< (Timed_Event const &timedEvent1) const
    { return _timeout < timedEvent1._timeout; }
  BOOLEAN operator> (Timed_Event const &timedEvent1) const
    { return _timeout > timedEvent1._timeout; }

  MSecs getTimeout (void) const { return _timeout; }
  void sendSignal (void) const { Event::sendSignal(getNode()); }

 private:
  MSecs _timeout;
};

class Event_List : public list<Event>
{
 public:
  unsigned int count(Signal_Enum signal) const ;

  void insert(Event const &event) { insertFirst(event); }
};

class State_Event_List : public list<State_Event>
{
 public:
  void insert(State_Event const &stateEvent) { insertFirst(stateEvent); }
};

typedef Const_List_Iterate<Event> Const_Event_Iterator;
typedef List_Iterate<Event> Event_Iterator;

typedef Const_List_Iterate<State_Event> Const_State_Event_Iterator;
typedef List_Iterate<State_Event> State_Event_Iterator;

#ifdef macintosh
class Timed_Event_List : public list<Timed_Event> {};
#endif

#endif // INCevent
