/*****************************************************************************
 * 
 * PROJECT: Task Control Management.
 *
 * (c) Copyright 1996 Reid Simmons.  All rights reserved.
 *
 * FILE: tcm.h
 *
 * ABSTRACT: Public include file for the Task Control Management library.
 *           Defines the TCM API.
 *
 * $Source: /afs/cs.cmu.edu/project/TCA/Master/tcaV8/tcm/src/tcmPriv.h,v $ 
 * $Revision: 1.22 $
 * $Date: 2000/07/05 23:14:08 $
 * $Author: da0g $
 *
 * REVISION HISTORY:
 *
 * $Log: tcmPriv.h,v $
 * Revision 1.22  2000/07/05 23:14:08  da0g
 * Fixed delete of const pointer bug.
 *
 * Revision 1.21  2000/02/03 14:41:25  reids
 * Fixed improper templating bugs with tcmHandle's refIncr and refDecr (da0g)
 *
 * Revision 1.20  2000/01/19 21:26:43  reids
 * Added two new top-level functions:
 *   TCM_IsDoneHandling(ref) -- returns TRUE if the ref has raised success or
 * 			     failure.
 *   TCM_IsPostponed(ref) -- returns TRUE if the ref's action has finished
 * 		          executing, but the node has not yet raised
 * 			  success or failure.
 *
 * Revision 1.19  1999/08/05 17:22:37  reids
 * Changes needed because the API to the commInterface library was updated
 *   to be more generally applicable.  Removed all dependencies (at least in
 *   theory) on IPC.
 *
 * Revision 1.18  1999/08/04 14:00:19  reids
 * Distributed version of TCM.  Provides the ability for one process to
 *  add tasks, constraints, exception handlers, etc. with another process.
 *  The distributed version is created by compiling with -DDISTRIBUTED flag,
 *   which can be gotten by "gmaking" with DISTRIBUTED=1.
 *
 * Revision 1.17  1999/06/06 13:48:10  reids
 * Changes to TCM 1.3.0:
 *   New list data structures.
 *   New, generalized logging interface.
 *   Fixed firstChild to work if there are no children.
 *   Fixed code to deallocate a node.
 *   In certain circmumstances, don't send termination signal when task
 *     itself terminating.
 *
 * Revision 1.16  1999/03/02 01:31:14  da0g
 * Updated TCM_FailureData to use getNodeRefCountData
 *
 * Revision 1.15  1999/02/26 18:05:42  da0g
 * Modified Agenda::clearQueues not to exit until timePoint/timeout has passed.
 *
 * Revision 1.14  98/12/21  17:56:13  reids
 * Added TCM_GetAction.
 * 
 * Revision 1.13  1998/12/16 03:08:21  reids
 * Added support for "on termination" functions.
 *   Also enabled tca.h and tcm.h to co-exist (needed to change values
 *   of several constants).
 *
 * Revision 1.12  98/09/15  18:45:24  da0g
 * Enhanced exceptions to support multiple-name resolution and Ref_Count (automatically-destroyed) Data.
 * 
 * Revision 1.11  1998/08/05 10:57:30  reids
 * Made a compiler option for printing out debugging information.
 *   Created macros for TCM_TerminateAtAfter and TCM_DelayUntilAfter, for
 *   backwards compatibility.
 *
 * Revision 1.10  98/07/14  17:21:31  reids
 * Fixed a bug in handling Activate_Signal.
 * Changed TCM_DelayUntilAfter to TCM_DelayForAfter, changed TCM_
 *   TerminateAtAfter to TCM_TerminateInAfter to TCM_DelayForAfter, and
 *   changed order of arguments.
 * Added TCM_ActivateInAfter.
 * 
 * Revision 1.9  98/06/02  10:39:46  reids
 * Increased overall efficiency of TCM.
 * 
 * Revision 1.8  98/05/30  12:00:34  reids
 * Fixed "waitFor" if timepoint is achieved but there are still events on
 *   the timer queue.
 * Clear timeouts related to nodes that are completely achieved.
 * 
 * Revision 1.7  98/04/21  12:46:15  reids
 * Added InvokeWhen/InvokeAfter functionality, to invoke a function when a
 *   particular event occurs.
 * Added the DelayUntilAfter/TerminateAtAfter functionality, to delay/terminate
 *   a node by waiting some msecs after a particular event has occurred.
 * 
 * Revision 1.6  98/03/06  12:42:37  reids
 * Modifications made to support Solaris.
 * 
 * Revision 1.5  98/01/30  14:51:07  reids
 * Updated to compile under gcc 2.7.2 and under Linux.
 * Also, made STRING "const char *" and changed API to take const arguments,
 *   where applicable.
 * 
 * Revision 1.4  97/12/30  12:29:06  reids
 * Added option to *not* wait for timeouts and external events.
 *   Added flag to indicate whether node raised an exception (ie, failed).
 *   Added a "timer" monitor, which is like a polling monitor, except when
 *     activated it just invokes a function (rather than adding a new node)
 * 
 * Revision 1.3  97/12/29  17:06:36  reids
 * Version that has the basic functionality needed to support TDL.
 * 
 * Revision 1.2  97/12/04  17:50:30  reids
 * Another fairly stable version (except that monitors do not quite work)
 * 
 * Revision 1.1  97/11/21  14:06:48  reids
 * First release of TCM -- seems to be a stable version
 *
 *****************************************************************************/

#ifndef INCtcmPriv
#define INCtcmPriv

#define TCM_VERSION_MAJOR  2
#define TCM_VERSION_MINOR  1
#define TCM_VERSION_MICRO  0
#define TCM_VERSION_DATE "Jul-5-00"
#define TCM_COMMIT_DATE "$Date: 2000/07/05 23:14:08 $"

class Named_Object
{
 public:
  Named_Object() { name = NULL; }
  Named_Object(STRING theName) { name = strdup(theName); }
  Named_Object ( Named_Object const & obj )
	   { name = (obj.name == NULL) ? ((STRING) NULL) : strdup(obj.name);  }
  ~Named_Object() { delete[] (char *)name; }

  Named_Object &operator=(Named_Object const &obj)
    { if (name) delete[] (char *)name; name = strdup(obj.name); return *this; }
  BOOLEAN operator==(Named_Object const &obj) const
    { return streq(name, obj.name); }
  BOOLEAN operator< (Named_Object const &obj) const
    { return strcmp(name, obj.name) < 0; }
  STRING getName(void) const { return name; }

 protected:
  STRING name;
};

typedef tcmHandle<class Task_Tree_Node> Task_Tree_Ref;

inline BOOLEAN validRef(Task_Tree_Ref const &ref) { return *ref != NULL; }

#endif /* INCtcmPriv */
