/**************************************************************************
 * 
 * PROJECT: Task Control Management.
 *
 * (c) Copyright 1999 Reid Simmons.  All rights reserved.
 *
 * FILE: virtual.h
 *
 * ABSTRACT: Defines "virtual" task tree nodes used for distributing a
 *           single task tree amongst multiple processes.  A "virtual" node
 *           has "real" counterpart in another process.  "Virtual" nodes 
 *           know how to send messages to their real counterparts and how
 *           to interpret messages sent to them.
 *
 *   Example:      Process 1      Process 2
 *                   [A]             [A']
 *                   / \               \
 *                 [B] [C']            [C]
 *
 *       In process 1, nodes A and B are real, and C' is virtual;
 *       In process 2, node C is real and node A' is virtual.  Virtual
 *       Virtual node C' corresponds to real node C, and A' corresponds to A.
 *
 * $Revision: 1.2 $
 * $Date: 2000/01/19 21:26:43 $
 * $Author: reids $
 *
 * REVISION HISTORY:
 *
 * $Log: virtual.h,v $
 * Revision 1.2  2000/01/19 21:26:43  reids
 * Added two new top-level functions:
 *   TCM_IsDoneHandling(ref) -- returns TRUE if the ref has raised success or
 * 			     failure.
 *   TCM_IsPostponed(ref) -- returns TRUE if the ref's action has finished
 * 		          executing, but the node has not yet raised
 * 			  success or failure.
 *
 * Revision 1.1  1999/08/04 14:00:20  reids
 * Distributed version of TCM.  Provides the ability for one process to
 *  add tasks, constraints, exception handlers, etc. with another process.
 *  The distributed version is created by compiling with -DDISTRIBUTED flag,
 *   which can be gotten by "gmaking" with DISTRIBUTED=1.
 *
 **************************************************************************/

#ifndef INCvirtual
#define INCvirtual

#include "taskTree.h"

typedef class Virtual_Node *Virtual_Node_Ptr;

class Virtual_Node : public Task_Tree_Node
{
 public:
  Virtual_Node (STRING taskName, STRING host, STRING creator)
    : Task_Tree_Node(taskName)
    { setVirtualAddress(host, creator, getNodeId()); _initialize(); }
  Virtual_Node (STRING taskName, const Virtual_Address &vaddress)
    : Task_Tree_Node(taskName)
    { setVirtualAddress(vaddress); _initialize(); }
  ~Virtual_Node () { 
    fprintf(stderr, "Killing %s (%s/%s/%d)\n", getName(), _vaddress.host,
	    _vaddress.creator, _vaddress.id); }

  virtual BOOLEAN isGoal(void)      { tcmError("isGoal\n"); return FALSE; }
  virtual BOOLEAN isException(void) { tcmError("isException\n");return FALSE; }
  virtual BOOLEAN isMonitor(void)   { tcmError("isMonitor\n"); return FALSE; }

  virtual BOOLEAN validState(State_Enum state) const 
    { tcmError("validState\n"); return FALSE; }
  virtual BOOLEAN    isCurrentState(State_Enum state) const 
    { tcmError("isCurrentState\n"); return FALSE; }
  virtual State_Enum theCurrentState(State_Enum state) const
    { tcmError("theCurrentState\n"); return Unknown_State; }
  virtual BOOLEAN    isChild(Task_Tree_Ref const &childNode) const 
    { tcmError("isChild\n"); return FALSE; }

  virtual void addParent(Task_Tree_Ref const &parentNode,
			 BOOLEAN canBeLastChild, BOOLEAN addConstraints);
  virtual void addChild(Task_Tree_Ref const &childNode, BOOLEAN addConstraints);
  virtual void removeChild(Task_Tree_Ref const &childNode);

  virtual  Task_Tree_Node_Ptr firstChild(void) const
    { tcmError("firstChild\n"); return NULL; }
  virtual  Task_Tree_Node_Ptr lastChild(BOOLEAN ignoreVeryLast) const
    { tcmError("lastChild\n"); return NULL; }
  virtual  Task_Tree_Node_Ptr previousChild(Task_Tree_Ref const &child) const
    { tcmError("previousChild\n"); return NULL; }
  virtual  Task_Tree_Node_Ptr nextChild(Task_Tree_Ref const &child) const
    { tcmError("nextChild\n"); return NULL; }
  virtual  Task_Tree_Node_Ptr childNamed(STRING const name) const
    { tcmError("childNamed\n"); return NULL; }
  virtual  Task_Tree_Node_Ptr ancestorNamed(STRING const name) const
    { tcmError("ancestorNamed\n"); return NULL; }

  virtual  void execute (void) { tcmError("execute\n"); }
  virtual  void signal(const Event &event);

  virtual  void addExpectedEvent(Signal_Enum signal,
				 Task_Tree_Ref const &signallingNode,
				 const void *timerData = NULL);
  virtual  void addRequestedEvent(State_Enum state, Signal_Enum signal,
				  Task_Tree_Ref const &signalledNode);
  virtual  void addRequestedEvent(State_Enum state, 
				  SIMPLE_CALLBACK_FN_TYPE callbackFn, 
				  const void *data);
  void          invokeCallback (State_Enum state);

  virtual  void removeExpectedEvent(Signal_Enum signal,
				    Task_Tree_Ref const &signallingNode);
  virtual  void removeRequestedEvents(Signal_Enum signal,
				      Task_Tree_Ref const &signalledNode);
  virtual  void processRequestedEvents(void)
    { tcmError("processRequestedEvents\n"); }
  virtual  void transitionTo(State_Enum newState) { tcmError("transitionTo\n"); }
  virtual  void doneHandling(void) { tcmError("doneHandling\n"); }

  virtual  void setFailed (void) { tcmError("setFailed\n"); }

  // Return TRUE if failed, FALSE otherwise
  virtual  BOOLEAN getNodeFailed (void) const 
    { tcmError("getNodeFailed\n"); return FALSE; }
  // Return TRUE if any node of the tree failed, FALSE otherwise
  virtual  BOOLEAN getTreeFailed (void) const 
    { tcmError("getTreeFailed\n"); return FALSE; }

  virtual  BOOLEAN terminate (void);
  virtual  BOOLEAN deallocate (void);
  virtual  void suspend (void);
  virtual  void unsuspend (void);
  virtual  BOOLEAN isSuspended (void) const 
    { tcmError("isSuspended\n"); return FALSE; }

  virtual  BOOLEAN isDoneHandling (void) const 
    { tcmError("isDoneHandling\n"); return FALSE; }
  virtual  BOOLEAN isPostponed (void) const
    { tcmError("isPostponed\n"); return FALSE; }

  virtual void delayTermination  (void);
  virtual void enableTermination (void);
  virtual BOOLEAN isTerminationDelayed (void) const 
    { tcmError("isTerminationDelayed\n"); return FALSE; }    

  virtual STRING className(void) const { return "Virtual"; }

  virtual void setNodeData (void const *nodeData) { tcmError("setNodeData\n");}
  virtual void const *getNodeData (void) const
    { tcmError("getNodeData\n"); return NULL; }

  virtual Ref_Count * getNodeRefCountData() const
    { tcmError("getNodeRefCountData\n"); return NULL; }
  virtual void setNodeRefCountData ( Ref_Count * theNodeRefCountData )
    { tcmError("setNodeRefCountData\n");}

  virtual void setAction (TCM_Action_Ref action) { tcmError("setAction\n"); }
  virtual TCM_Action_Ref getAction (void) const 
    { tcmError("getAction\n"); return NULL; }

  virtual  void displayTree (FILE *stream, int position = 0) const;

  virtual  void addTerminationAction (const TCM_Task_Tree_Ref &terminationRef);

  virtual void findAndApplyExceptionHandler (Task_Tree_Ref parentNode,
					     Task_Tree_Ref failureNode,
					     Exception_Instance *exception,
					     const void *data,
					     BOOLEAN isRefCountFailureData);
  virtual void noExceptionHandler (STRING failure) const;

  virtual void addTimedEventAfterSignal (MSecs waitTime,
					 const Task_Tree_Ref nodeRef, 
					 Signal_Enum signal,
					 State_Enum state);

 protected: // Functions
  virtual  inline BOOLEAN _noneExpected(Signal_Enum signal) const 
    { tcmError("_noneExpected\n"); return FALSE; }
  // Different classes of nodes may handle the same signal in different ways.
  virtual void _handleSignal (Signal_Enum signal)
    { tcmError("_handleSignal\n"); }
  virtual void _intraNodeConstraints (void) 
    { tcmError("_intraNodeConstraints\n"); }

 private: // Functions
   // Initialize task tree node data structure
  virtual  void _initialize(void);
  virtual  void _processSignal (Signal_Enum signal) 
    { tcmError("_processSignal\n"); }

  virtual  BOOLEAN _handleTransition (State_Enum newState)
    { tcmError("_handleTransition\n"); return FALSE; }
  virtual  void _startTerminating (void) { tcmError("_startTerminating\n"); }
  virtual  void _destroy (void) { tcmError("_destroy\n"); }

  //  Unsuspend a single node, and trigger any waiting transitions
  virtual  void _unsuspendNode (void) { tcmError("_unsuspendNode\n"); }
  
  virtual  STRING _nodeStatusString (void) const 
    { tcmError("_nodeStatusString\n"); return "Unknown"; }

  virtual void _forceTimeouts (void) { tcmError("_forceTimeouts\n"); }
};


#endif // INCvirtual
